/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */
package com.indexdata.torus2;

import java.util.Date;
import com.indexdata.torus2.exception.InternalIOException;
import com.indexdata.torus2.exception.NotFoundException;
import com.indexdata.torus2.exception.MalformedInputException;
import com.indexdata.torus2.query.QueryException;
import com.indexdata.torus2.query.RecordComparator;
import com.indexdata.torus2.model.ParentRealm;
import com.indexdata.torus2.model.ResultSet;
import com.indexdata.torus2.model.RealmDefinition;
import com.indexdata.torus2.util.Pair;
import com.indexdata.torus2.model.Record;
import com.indexdata.torus2.model.Record.Layer;
import com.indexdata.torus2.model.Record.Layer.Field;
import java.util.Map;
import com.indexdata.torus2.model.Facets;
import com.indexdata.torus2.model.Facets.Facet;
import java.util.Iterator;
import com.indexdata.torus2.rw.RealmRWFactory;
import com.indexdata.torus2.rw.RealmReaderWriter;
import java.util.Collections;
import org.z3950.zing.cql.CQLSortNode;
import com.indexdata.masterkey.config.MasterkeyConfiguration;
import com.indexdata.torus2.query.match.LayerMatcher;
import com.indexdata.torus2.query.MatcherBuilder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import org.apache.log4j.Logger;
import org.z3950.zing.cql.CQLNode;

import static com.indexdata.torus2.query.MatchingUtil.parseQuery;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import org.apache.commons.httpclient.util.DateUtil;
/**
 *
 * @author jakub
 */
public class Realm {
  private String host;
  private Logger logger = Logger.getLogger("com.indexdata.torus2");
  private RealmReaderWriter rw;
  private RealmDefinition rd;
  private Date startupDate;

  private Layer propagateRecordUpdate(Layer override) throws InternalIOException {
    //propagate updates to inherriting realms
    if (rw.isCachedAsParent()) {
      //construct final, cache as original
      logger.info("Updating record '"+override.getId()+"' in the parent cache.");
      Layer original = null;
      if (override.getWorldId() != null && !override.getWorldId().isEmpty()) {
        try {
          original = rw.readWorldRecord(override.getWorldId());
        } catch (MalformedInputException ex) {
          throw new InternalIOException(ex);
        }
      }
      Layer utmost = mergeLayers(original, override);
      utmost.setName("original");
      utmost.setRealmName("world");
      utmost.setWorldId(null);
      rw.updateCachedAsOriginal(utmost);
      return utmost;
    }
    return null;
  }
  
  private void propagateParentUpdate(String parentId) throws InternalIOException {
    if (rw.isCachedAsParent()) {
      // read through this realm records, update cache
      logger.info("Updating caches following parent update...");
      List<Layer> overrides = rw.readOverrideRecords();
      for (Layer override : overrides) {
        String worldId = override.getWorldId();
        String prId = worldId != null 
          ? worldId.substring(0, worldId.indexOf('.'))
          : null;
        if (parentId.equals(prId)) {
          // this record might have been updated/removed
          Layer original = null;
          try {
            original = rw.readWorldRecord(worldId);
          } catch (InternalIOException rio) {
            logger.warn("World record '"+worldId+"' missing, possibly removed.");
          } catch (MalformedInputException mie) {
            logger.warn("Cannot read world record: malformed id: " + mie.getMessage());
          }
          if (original == null) {
            // possibly removed in the parent so remove from cache
            try {
              rw.deleteCachedAsOriginal(override.getId());
            } catch (InternalIOException rio) {
              logger.warn("Previous instance of '"+override.getId()+
                "' not found in parent cache, possibly an orphan");
            }
          } else {
            // possibly updated, refresh in the cache
            Layer utmost = mergeLayers(original, override);
            utmost.setName("original");
            utmost.setRealmName("world");
            utmost.setWorldId(null);
            rw.updateCachedAsOriginal(utmost);
          }
        }
      }
    }
  }

  private List<Record> readOrFetchWorldRecords() throws InternalIOException {
    Map<String,LinkedList<Pair<Layer,ParentRealm>>> dedupe =
      new HashMap<String, LinkedList<Pair<Layer,ParentRealm>>>();
    List<Record> allOrigs = new ArrayList<Record>();
    List<ParentRealm> parents = rw.readParentRealms();
    if (parents.size() <= 0) return allOrigs; //empty list
    ExecutorService pool = Executors.newFixedThreadPool(parents.size());
    List<Future<List<Layer>>> blocking = new ArrayList<Future<List<Layer>>>(parents.size());
    for (ParentRealm pr : parents) {
      Pair<List<Layer>, Date> cached = rw.readWorldRecordsForParent(pr.getId());
      Date lastRefreshed = cached != null ? cached.getSecond() : null;
      if (cached == null || pr.needsRefresh(lastRefreshed)) {
        //we don't have them locally, we'll try to fetch and store them
        String parentUrl = pr.resolveUrl(host).toString();
        RealmClient rc = new RealmClient();
        if (cached == null) {
          logger.warn("Records for parent " + pr.getName() + "<" + parentUrl +"> "
              + "missing (torus booting?), attempting retrieval from remote..");
          Callable<List<Layer>> request  = rc.fetchAndCacheRecordsAsync(parentUrl, 0, pr, rw, 
            lastModified(lastRefreshed, startupDate));
          Future<List<Layer>> response = pool.submit(request);
          blocking.add(response);
        } else {
          logger.info("Records for parent " + pr.getName() + "<" + parentUrl +"> "
              + "outdated (last refreshed at "+DateUtil.formatDate(lastRefreshed)
            +"), using cached version and scheduling background update from remote..");
          Callable<List<Layer>> request = rc.fetchAndCacheRecordsAsync(parentUrl, 0, pr, rw,
            lastModified(lastRefreshed, startupDate));
          pool.submit(request);
        }
      }
    }
    //wait for missing cache records
    for (Future<List<Layer>> req : blocking) {
      try {
        List<Layer> recs = req.get(); //never-null, already cached
      } catch (ExecutionException ee) {
        if (ee.getCause() instanceof InternalIOException)
          throw (InternalIOException) ee.getCause();
        else //should never happen
          throw new InternalIOException(
            "Unexpected error when retrieving records from parent", 
            ee.getCause());
      } catch (InterruptedException ie) {
        throw new InternalIOException(
          "Unexpected error when retrieving records from parent (interrupted)", 
          ie);
      }
    }
    for (ParentRealm pr : parents) {
      Pair<List<Layer>,Date> cached = rw.readWorldRecordsForParent(pr.getId());
      List<Layer> realmOrigs = cached.getFirst();
      if (rd.getMatchKey() != null) {
        for (Layer l : realmOrigs) {
          String matchKey = rd.getMatchKey().generate(l);
          logger.debug("Generated matchKey '" + matchKey + "' for '"
            +l.getId()+"'");
          LinkedList<Pair<Layer,ParentRealm>> prev = dedupe.get(matchKey);
          if (prev != null) {
            Pair<Layer,ParentRealm> head = prev.peekFirst();
            if (head.getSecond().getPriority() >= pr.getPriority()) {
              //both when grater or equal -- add as head
              prev.addFirst(new Pair<Layer,ParentRealm>(l, pr));
            } //do nothing otherwise
          } else {
            LinkedList<Pair<Layer,ParentRealm>> docList =
              new LinkedList<Pair<Layer,ParentRealm>>();
            docList.addFirst(new Pair<Layer,ParentRealm>(l, pr));
            dedupe.put(matchKey, docList);
          }
        }
      } else {
        for (Layer l : realmOrigs)
          allOrigs.add(new Record(rd.getRealmType(), pr.getId(), l));
      }
    }
    if (rd.getMatchKey() != null) {
      for (LinkedList<Pair<Layer,ParentRealm>> list : dedupe.values()) {
        int prevPrio = list.peekFirst().getSecond().getPriority();
        for (Pair<Layer,ParentRealm> p : list) {
          Layer l = p.getFirst();
          ParentRealm pr = p.getSecond();
          if (prevPrio < pr.getPriority()) break;
          allOrigs.add(new Record(rd.getRealmType(), pr.getId(), l));
        }
      }
    }
    return allOrigs;
  }

  private Record readOrFetchWorldRecord(String id)
    throws InternalIOException, MalformedInputException {
    Layer wRec = rw.readWorldRecord(id);
    String prId = id.substring(0, id.indexOf('.'));
    ParentRealm pr = rw.readParentRealm(prId);
    if (pr == null) throw new NotFoundException("Parent for '"+id+"' missing");
    if (wRec == null) {
      logger.warn("World record '"+id+"' missing or unreadable");
      logger.info("Retrieving world record '"+id+"' from parent..");
      RealmClient rc = new RealmClient();
      rc.fetchAndCacheRecords(pr.resolveUrl(host).toString(), 0, pr, rw, null);
      wRec = rw.readWorldRecord(id);
      if (wRec == null) {
        throw new InternalIOException("Missing cached records for "+pr.getName()+
          "<"+pr.resolveUrl(host).toString()+">");
      }
    }
    return new Record(rd.getRealmType(), prId, wRec);
  }

  public enum LayerType {
    FINAL("final"), OVERRIDE("override"), ORIGINAL("original"), MASTER_OVERRIDE("master-override");
    private String name;
    private LayerType(String name) {
      this.name = name;
    }
    public String getName() {
      return name;
    }
    public static List<LayerType> parse(String names) throws MalformedInputException {
      String[] layers = names.trim().split(",\\s*");
      List<LayerType> lt = new ArrayList<Realm.LayerType>();
      for (String layer : layers) {
        try {
          lt.add(LayerType.valueOf(layer.toUpperCase().replace('-', '_')));
        } catch (IllegalArgumentException iae) {
          throw new MalformedInputException("Unknown layer type '" + layer + "'");
        }
      }
      return lt;
    }
  };

  protected Realm(RealmReaderWriter rw, RealmDefinition rd, String host, Date startupDate) {
    this.rw = rw;
    this.rd = rd;
    this.host = host;
    this.startupDate = startupDate;
  }

  public static Realm createRealm(RealmDefinition rd, MasterkeyConfiguration cfg,
    String host, String proto, Date startupDate)
    throws InternalIOException {
    RealmReaderWriter rw = RealmRWFactory.getRealmRW(rd.getRealmName(), cfg, 
      host, proto);
    return rw.create(rd) ? new Realm(rw, rd, host, startupDate) : null;
  }

  public static Realm openRealm(String realmName, MasterkeyConfiguration cfg,
    String host, String proto, Date startupDate) throws InternalIOException {
    RealmReaderWriter rw = RealmRWFactory.getRealmRW(realmName, cfg, host, 
      proto);
    return rw.open()
      ? new Realm(rw, rw.describe(), host, startupDate)
      : new Realm(rw, null, host, startupDate);
  }

  public static void destroyRealm(String realmName, MasterkeyConfiguration cfg,
    String host, String proto) throws InternalIOException {
    RealmReaderWriter rw = RealmRWFactory.getRealmRW(realmName, cfg, host, 
      proto);
    rw.destroy();
  }
  
  public static Realm updateRealm(RealmDefinition rd, MasterkeyConfiguration cfg,
    String host, String proto, Date startupDate) throws InternalIOException {
    RealmReaderWriter rw = RealmRWFactory.getRealmRW(rd.getRealmName(), cfg, host, 
      proto);
    Realm updated = rw.update(rd) ? new Realm(rw, rd, host, startupDate): null;
    return updated;
  }

  public RealmDefinition getDefinition() {
    return rd;
  }

  public Record getWorldRecord(String id)
    throws InternalIOException, MalformedInputException {
    Record r = readOrFetchWorldRecord(id);
    r.setLastModified(r.getLayer("original").getLastModified());
    return r;
  }

  public ResultSet getWorldRecords(CQLNode query, List<String> facetNames,
    int recursive)
    throws InternalIOException, QueryException {
    if (recursive-- > 0) {
      List<ParentRealm> parents = rw.readParentRealms();
      for (ParentRealm pr : parents) {
        RealmClient rc = new RealmClient();
        rc.fetchAndCacheRecords(pr.resolveUrl(host).toString(), 0, pr, rw, null);
      } 
    }
    List<Record> originals = readOrFetchWorldRecords();
    List<Record> matchRecs = null;
    CQLSortNode sortBy = null;
    if (query instanceof CQLSortNode) {
      // pick out criteria part from sort node
      sortBy = (CQLSortNode) query;
      query = sortBy.getSubtree();
    }
    LayerMatcher m = MatcherBuilder.build(query);
    matchRecs = new ArrayList<Record>(originals.size());
    Facets facets = facetNames != null && !facetNames.isEmpty() ?
      new Facets(facetNames) : null;
    Date prevDate = null;
    for (Record origRec : originals) {
      Layer original = origRec.getLayer("original");
      if (m == null || m.match(original)) {
        matchRecs.add(origRec);
        updateFacets(facets, original);
        Date nextDate = original.getLastModified();
        if (prevDate == null || (nextDate != null && nextDate.after(prevDate)))
          prevDate = nextDate;
      }
    }
    if (sortBy != null) {
      Collections.sort(matchRecs, new RecordComparator(sortBy, "original"));
    }
    return new ResultSet(matchRecs, facets, -1, -1, prevDate);
  }
  
  /**
   * Tries to retrieve ParentRealm from a prefixed id.
   * @param id
   * @return
   * @throws InternalIOException 
   */
  private ParentRealm getParentFromRecId(String id) throws InternalIOException {
    int idx = id.indexOf(".");
    if (idx <= 0) return null;
    String prId = id.substring(0, idx);
    return rw.readParentRealm(prId);
  }
  
  private boolean isAutoInherit(String id) throws InternalIOException {
    ParentRealm pr = getParentFromRecId(id);
    return pr == null ? false : isAutoInherit(pr);
  }
  
  private boolean isAutoInherit(ParentRealm pr) {
    return pr.getAutoInherit().equals(ParentRealm.Trilean.UNDEFINED)
      ? rd.isAutoInherit()
      : pr.getAutoInherit().equals(ParentRealm.Trilean.TRUE);
  }

  public Record getRealmRecord(String id, List<LayerType> layerNames) throws
    InternalIOException {
    if (layerNames.size() < 1) throw new IllegalArgumentException("Specify at least one layer.");
    //we need override always and original sometimes
    Layer override = null;
    try {
      override = rw.readOverrideRecord(id);
    } catch (NotFoundException nfe) {
      //rethrow if this is not an auto-inherit record
      if (!isAutoInherit(id)) throw nfe; 
    }
    Layer original = null;
    Record r = null;
    String worldId = override != null ? override.getWorldId() : id;
    if (worldId != null) {
      //we could re-use the instance but we would need to hide the layer
      Record origRec;
      try {
        origRec = readOrFetchWorldRecord(worldId);
      } catch (MalformedInputException ex) {
        throw new InternalIOException(ex);
      }
      original = origRec.getLayer("original");
      r = new Record(rd.getRealmType(), origRec.getParentId());
    } else {
      r = new Record(rd.getRealmType(), null);
    }
    //create auto-inherit override layer
    if (override == null) {
        override = new Layer("override");
        override.setId(id);
        override.setRealmName(rd.getRealmName());
        override.setDisabled(false);
        override.setWorldId(worldId);
    }
    Layer merged = mergeLayers(original, override);
    //check the masters
    List<Layer> masters = safeReadMasters();
    List<Layer> matchedMasters = checkMasters(masters, merged);
    //re-merge if any masters
    if (matchedMasters != null && matchedMasters.size()>0) {
      merged = mergeWithMasters(worldId, original, override, matchedMasters);
    }
    //record last-modified does not depend on layers requested
    r.setLastModified(merged.getLastModified());
    for (LayerType lt : layerNames) {
      switch (lt) {
        case FINAL: {
          r.addLayer(merged);
          break;
        }
        case OVERRIDE: {
          r.addLayer(override);
          break;
        }
        case ORIGINAL: {
          if (original != null)
            r.addLayer(original);
          break;
        }
        case MASTER_OVERRIDE: {
          if (matchedMasters != null) {
            for (Layer master : matchedMasters) {
              r.addLayer(master);
            }
          }
          break;
        }
        default:
          throw new IllegalArgumentException("Unknown layer type");
      }
    }
    return r;
  }
  
  private List<Layer> safeReadMasters() throws InternalIOException {
    try {
      return rw.readMasterRecords();
    } catch (NotFoundException nfe) {
      //old torus2 may be missing masters dir
      logger.warn("Realm '"+rd.getRealmName()+"' is missing 'masters' dir");
    }
    return null;
  }
  
  private List<Layer> checkMasters(List<Layer> masters, Layer merged) 
    throws InternalIOException {
    if (masters == null) return null;
    List<Layer> matchedMasters = new ArrayList<Layer>(masters.size());
    for (Layer master : masters) {
      try {
        CQLNode matchQuery = parseQuery(master.getMatchQuery());
        LayerMatcher mm = MatcherBuilder.build(matchQuery);
        if (mm == null || mm.match(merged))
          matchedMasters.add(master);
      } catch (QueryException qe) {
        throw new InternalIOException("Bad matchQuery '"
          + master.getMatchQuery()
          + "' in master record '" + master.getId() + "'", qe);
      }
    }
    return matchedMasters;
  }
  
  private Layer mergeWithMasters(String parentId, Layer original, Layer override,
    List<Layer> matchedMasters) {
    Layer[] all = new Layer[matchedMasters.size() + 2];
    all[0] = original;
    for (int i = 0; i < matchedMasters.size(); i++) {
      all[i + 1] = matchedMasters.get(i);
    }
    all[all.length - 1] = override;
    Layer result = mergeLayers(all);
    //we don't want master id to overlay original id, set to expected id
    if (override == null && parentId != null)
      result.setId(parentId);
    return result;
  }

  public ResultSet getRealmRecords(CQLNode query, List<LayerType> layerNames, 
    List<String> facetNames)
    throws InternalIOException, QueryException {
    return getAllRecordsInternal(query, layerNames, facetNames, false);
  }

  public Record updateRealmRecord(String id, Record update)
    throws InternalIOException, MalformedInputException {
    Layer after = update.getLayer("override");
    if (after == null) throw new MalformedInputException("Record must contain 'override' layer");
    //auto-inherit magic
    Layer before = null;
    try {
      before = rw.readOverrideRecord(id);
    } catch (NotFoundException nfe) {
      if (!isAutoInherit(id)) throw nfe;
      try {
        Record origRec = readOrFetchWorldRecord(id);
        //at this point we are simply creating an override
        after.setWorldId(id);
        return createRealmRecord(update);
      } catch (NotFoundException nfe2) {
        //rethrow previous nfe
        throw nfe;
      }
    }
    Layer updated = updateLayer(before, after);
    updated.setLastModified(new Date());
    updated = rw.writeOverrideRecord(id, updated);
    propagateRecordUpdate(updated);
    return update.setLayer(updated);
  }

  public Record createRealmRecord(Record initial) throws InternalIOException,
    MalformedInputException {
    Layer override = initial.getLayer("override");
    if (override == null) throw new MalformedInputException("Record must contain 'override' layer");
    override.setLastModified(new Date());
    override.setCreationDate(new Date());
    override = rw.writeOverrideRecord(override);
    propagateRecordUpdate(override);
    return initial.setLayer(override);
  }

  public void deleteRealmRecord(String id) throws InternalIOException {
    rw.deleteOverrideRecord(id);
    if (rw.isCachedAsParent())
      rw.deleteCachedAsOriginal(id);
  }

  public List<ParentRealm> getParentRealms() throws InternalIOException {
    List<ParentRealm> prs = rw.readParentRealms();
    for (ParentRealm pr : prs) {
      Pair<List<Layer>, Date> cached = rw.readWorldRecordsForParent(pr.getId());
      if (cached != null) {
        pr.setLastRefreshed(cached.getSecond());
      }
    }
    return prs;
  }

  public ParentRealm getParentRealm(String prId) throws InternalIOException {
    ParentRealm pr = rw.readParentRealm(prId);
    //set lastRefreshed
    Pair<List<Layer>, Date> cached = rw.readWorldRecordsForParent(prId);
    if (cached != null) {
      pr.setLastRefreshed(cached.getSecond());
    }
    return pr;
  }

  public ParentRealm createParentRealm(ParentRealm pr) throws InternalIOException {
    RealmClient rc = new RealmClient();
    //try to connect first
    //fetchAndCacheRecords will notice that the parent does not extist and create it
    rc.fetchAndCacheRecords(pr.resolveUrl(host).toString(), 0, pr, rw, null);
    return pr;
  }

  public ParentRealm updateParentRealm(String prId, ParentRealm pr)
    throws InternalIOException {
    RealmClient rc = new RealmClient();
    rc.fetchAndCacheRecords(pr.resolveUrl(host).toString(), 0, pr, rw, null);
    pr.setLastUpdated(new Date());
    pr = rw.writeParentRealm(pr.getId(), pr);
    propagateParentUpdate(prId);
    return pr;
  }

  public void deleteParentRealm(String prId)
    throws InternalIOException {
    rw.deleteWorldRecordsByParentId(prId);
    rw.deleteParentRealm(prId);
    propagateParentUpdate(prId);
  }

  private void updateFacets(Facets facets, Layer layer) {
    if (facets != null) {
      for (Facet f : facets.getFacets()) {
        for (Field field : layer) {
          if (field.getName().equals(f.getName())) {
            if (f.getSplitRegex() != null) {
              String[] items = f.getSplitRegex().split(field.getContents());
              for (String item : items) {
                f.addTerm(item);
              }
            } else {
              f.addTerm(field.getContents());
            }
          }
        }
      }
    }
  }
  
  public ResultSet getAllRecords(CQLNode query, List<LayerType> layerNames, 
    List<String> facetNames) throws InternalIOException, QueryException {
    return getAllRecordsInternal(query, layerNames, facetNames, true);
  }

  private ResultSet getAllRecordsInternal(CQLNode query, List<LayerType> layerNames, 
    List<String> facetNames, boolean isAdmin)
    throws InternalIOException, QueryException {
    if (layerNames.size() < 1) throw new IllegalArgumentException("Specify at least one layer.");
    //do some early filtering
    CQLSortNode sortBy = null;
    if (query instanceof CQLSortNode) {
      // pick out criteria part from sort node
      sortBy = (CQLSortNode) query;
      query = sortBy.getSubtree();
    }
    LayerMatcher m = MatcherBuilder.build(query);
    List<Record> candidates = readOrFetchWorldRecords();
    //it might be a good idea to load that into a Map indexed by worldId
    List<Layer> overrides = rw.readOverrideRecords();
    //read the masters
    List<Layer> masters = safeReadMasters();
    List<Record> matchRecs = new ArrayList<Record>(candidates.size());
    Facets facets = facetNames != null && !facetNames.isEmpty() ?
      new Facets(facetNames) : null;
    int overidden = 0, local = 0;
    //first pick-up local
    Date prevDate = null;
    for (Iterator<Layer> it = overrides.iterator(); it.hasNext();) {
      Layer override = it.next();
      //handle purely-local
      String worldId = override.getWorldId();
      if (worldId == null) {
        //get rid of disabled records, non-admin view
        if (!isAdmin && override.isDisabled()) {
          it.remove();
          continue;
        }
        //check masters
        List<Layer> matchedMasters = checkMasters(masters, override);
        //re-merge if any masters
        boolean hasMaster = false;
        Layer merged;
        if (matchedMasters != null && matchedMasters.size()>0) {
          hasMaster = true;
          merged = mergeWithMasters(null, null, override, matchedMasters);
        } else {
          //dummy clone
          merged = override;
        }
        if (m == null || m.match(merged)) {
          Record r = new Record(rd.getRealmType(), null);
          matchRecs.add(r);
          updateFacets(facets, merged);
          boolean hasFinal = false;
          for (LayerType lt : layerNames) {
            switch (lt) {
              case OVERRIDE:
                r.addLayer(override);
                break;
              case ORIGINAL:
                //no-op
                break;
              case FINAL:
                hasFinal = true;
                //clone if dummy
                merged = hasMaster ? merged : mergeLayers(merged);
                r.addLayer(merged);
                break;
              case MASTER_OVERRIDE: {
                if (matchedMasters != null) {
                  for (Layer master : matchedMasters) {
                    r.addLayer(master);
                  }
                }
                break;
              }
              default:
                throw new IllegalStateException("Unknonw layer type");
            }
          }
          //put the final if not added and has master
          if (hasMaster && !hasFinal) {
            r.addLayer(merged);
            r.hideLayer("final");
          }
        }
        it.remove();
        local++;
        // last-modified
        Date nextDate = override.getLastModified();
        if (prevDate == null || (nextDate != null && nextDate.after(prevDate)))
          prevDate = nextDate;
      }
    }
    //then take world and corresponding parents
    for (Record originalRec : candidates) {
      Layer original = originalRec.getLayer("original");
      String expectedWorldId = originalRec.getParentId() + "." + original.getId();
      //pop override
      boolean hasOverrides = false;
      for (Iterator<Layer> it = overrides.iterator(); it.hasNext();) {
        Layer override = it.next();
        String worldId = override.getWorldId();
        if (worldId.equals(expectedWorldId)) {
          hasOverrides = true;
          //get rid of disabled records
          if (!isAdmin && override.isDisabled()) {
            it.remove();
            continue;
          }
          //TODO clone the layer?
          Layer merged = mergeLayers(original, override);
          //apply master before search
          List<Layer> matchedMasters = checkMasters(masters, merged);
          //re-merge if any masters
          if (matchedMasters != null && matchedMasters.size()>0) {
            merged = mergeWithMasters(null, original, override, matchedMasters);
          }
          if (m == null || m.match(merged)) {
            //check masters
            Record r = new Record(rd.getRealmType(), originalRec.getParentId());
            boolean hasFinal = false;
            for (LayerType lt : layerNames) {
              switch (lt) {
                case OVERRIDE:
                  r.addLayer(override);
                  break;
                case ORIGINAL:
                  r.addLayer(original);
                  break;
                case FINAL:
                  hasFinal = true;
                  r.addLayer(merged);
                  break;
                case MASTER_OVERRIDE: {
                  if (matchedMasters != null) {
                    for (Layer master : matchedMasters) {
                      r.addLayer(master);
                    }
                  }
                  break;
                }
                default:
                  throw new IllegalStateException("Unknonw layer type");
              }
            }
            if (!hasFinal) {
              r.addLayer(merged);
              r.hideLayer("final");
            }
            matchRecs.add(r);
            updateFacets(facets, merged);
            // date
            Date nextDate = merged.getLastModified();
            if (prevDate == null || (nextDate != null && nextDate.after(prevDate)))
              prevDate = nextDate;
          }
          it.remove();
          overidden++;
          //world rec can have more than one child, don't stop yet
        }
      }
      if ((isAdmin || isAutoInherit(expectedWorldId)) && !hasOverrides) {
        //so we deal with pure un-used world, still master applies
        //for auto-inherit we need to create override layer on the fly
        Layer override = null;
        if (isAutoInherit(expectedWorldId)) {
          override = new Layer("override");
          override.setId(expectedWorldId);
          override.setRealmName(rd.getRealmName());
          override.setDisabled(false);
          override.setWorldId(expectedWorldId);
        }
        Layer merged = mergeLayers(original, override);
        //check masters
        List<Layer> matchedMasters = checkMasters(masters, merged);
        if (matchedMasters != null && matchedMasters.size()>0) {
          merged = mergeWithMasters(expectedWorldId, original, override, matchedMasters);
        } 
        if (m == null || m.match(merged)) {
          Record r = new Record(rd.getRealmType(), originalRec.getParentId());
          matchRecs.add(r);
          updateFacets(facets, merged);
          boolean hasFinal = false;
          for (LayerType lt : layerNames) {
            switch (lt) {
              case OVERRIDE:
                if (override != null)
                  r.addLayer(override);
                break;
              case ORIGINAL:
                r.addLayer(original);
                break;
              case FINAL:
                hasFinal = true;
                r.addLayer(merged);
                break;
              case MASTER_OVERRIDE: {
                if (matchedMasters != null) {
                  for (Layer master : matchedMasters) {
                    r.addLayer(master);
                  }
                }
                break;
              }
              default:
                throw new IllegalStateException("Unknown layer type");
            }
          }
          //make sure final is include, even if not requested
          if (!hasFinal) {
            r.addLayer(merged);
            r.hideLayer("final");
          }
          Date nextDate = original.getLastModified();
          if (prevDate == null || (nextDate != null && nextDate.after(prevDate)))
          prevDate = nextDate;
        }
      }
    }
    //what remains are orphans
    if (logger.isDebugEnabled()) {
      for (Layer orphan : overrides) {
        logger.debug("Ignoring missing original '"+orphan.getWorldId()+
            "' ref by override '"+orphan.getId()+"'");
      }
    }
    if (overrides.size() > 0) {
      logger.warn("Total "+overrides.size()+" records ignored due to missing original, "
        + "see DEBUG for a complete list.");
    }
    
    if (sortBy != null) {
      //since we avoid cloning layers, need to 
      Collections.sort(matchRecs, new RecordComparator(sortBy, "final", "override", "original"));
    }
    return new ResultSet(matchRecs, facets, overidden, local, prevDate);
  }

  public List<RealmDefinition> listAllRealms() throws InternalIOException {
    return rw.listRealms();
  }
  
  //master record stuff
  
  public List<Record> listMasterRecords() throws InternalIOException {
    List<Layer> layers = rw.readMasterRecords();
    List<Record> records = new ArrayList<Record>(layers.size());
    for (Layer l : layers) {
      records.add(new Record("master", null, l));
    }
    return records;
  }
  
  public Record getMasterRecord(String id) throws InternalIOException {
    return new Record("master", null, rw.readMasterRecord(id));
  }
  
  public void deleteMasterRecord(String id) throws InternalIOException {
    rw.deleteMasterRecord(id);
    //propagate change
  }
  
  public Record createMasterRecord(Record master) throws InternalIOException, MalformedInputException {
    Layer masterOverride = master.getLayer("master-override");
    if (masterOverride == null)
      throw new MalformedInputException("Master record must contain 'master-overide' layer");
    Date now = new Date();
    masterOverride.setCreationDate(now);
    masterOverride.setLastModified(now);
    rw.writeMasterRecord(masterOverride);
    //propagate change
    return master;
  }
  
  public Record updateMasterRecord(String id, Record updateRec)
    throws InternalIOException, MalformedInputException {
    Layer before = rw.readMasterRecord(id);
    Layer update = updateRec.getLayer("master-override");
    if (update == null) throw new MalformedInputException("Record must contain 'master-override' layer");
    Layer after = updateLayer(before, update);
    after.setLastModified(new Date());
    //null values are used to notify no change
    if (update.getMatchQuery() != null)
      after.setMatchQuery(update.getMatchQuery());
    if (update.getLabel() != null)
      after.setLabel(update.getLabel());
    after = rw.writeMasterRecord(id, after);
    //propagate change
    return updateRec.setLayer(after);
  }
  
  //
  
  private Layer mergeLayers(Layer... override) {
    return mergeLayersOn(new Layer("final"), override);
  }
  
  private Layer mergeLayersOn(Layer merged, Layer... overrides) {
    Date lastModified = null;
    //perform overrides in order
    for (Layer override : overrides) {
      if (override == null) continue;
      //use top-most for fixed attrs and fields
      //null-safe
      merged.setId(override.getId());
      merged.setRealmName(override.getRealmName());
      merged.setWorldId(override.getWorldId());
      merged.setDisabled(override.isDisabled());
      for (Layer.Field f : override) {
        merged.setField(f.getName(), f.getContents());
      }
      //use last non-null for creation date
      if (override.getCreationDate() != null)
        merged.setCreationDate(override.getCreationDate());
      //use oldest for last modified date
      if (override.getLastModified() != null) {
        if (lastModified == null 
          || lastModified.before(override.getLastModified())) {
          lastModified = override.getLastModified();
          merged.setLastModified(lastModified);
        }
      }
    }
    return merged;
  }

  private Layer updateLayer(Layer before, Layer update) {
    Layer after = new Layer(before.getName());
    after.setId(before.getId());
    after.setRealmName(before.getRealmName());
    after.setCreationDate(before.getCreationDate());
    //should we allow overriding worldIds?
    after.setWorldId(before.getWorldId());
    after.setDisabled(update.isDisabled());
    //master stuff
    after.setMatchQuery(before.getMatchQuery());
    after.setLabel(before.getLabel());
    for (Layer.Field f : update) {
      if (!f.isGone())
        after.setField(f.getName(), f.getContents());
    }
    for (Layer.Field f : before) {
      Layer.Field ff = update.getField(f.getName());
      if (ff == null)
        after.setField(f.getName(), f.getContents());
    }
    return after;
  }
  
  private Date lastModified(Date lastRefreshed, Date startupDate) {
    return startupDate != null && lastRefreshed != null 
      && startupDate.before(lastRefreshed) 
      ? lastRefreshed
      : null;
  }
}
