/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */
package com.indexdata.torus2;

import com.indexdata.torus2.util.URIHelper;
import java.io.UnsupportedEncodingException;
import java.util.Date;
import com.indexdata.torus2.exception.InternalIOException;
import com.indexdata.torus2.exception.MalformedInputException;
import com.indexdata.torus2.model.ParentRealm;
import com.indexdata.torus2.model.Record.Layer;
import com.indexdata.torus2.rw.RealmReaderWriter;
import com.indexdata.utils.XmlUtils;
import com.sun.jersey.core.util.Base64;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import static com.indexdata.torus2.util.DOMHelper.*;
import com.indexdata.torus2.util.Pair;
import com.indexdata.utils.DateUtil;
import java.text.ParseException;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;

/**
 *
 * @author jakub
 */
public class RealmClient {
  private Logger logger = Logger.getLogger("com.indexdata.torus2");
  private static ConcurrentHashMap<String, String> ongoingRequest = new ConcurrentHashMap<String, String>();
  private final int ONGOING_WAIT_TOTAL = 30000;
  private final int ONGOIGN_WAIT_CHECK = 1000;

  public class RealmClientException extends InternalIOException {
    /**
     * 
     */
    private static final long serialVersionUID = 1461182381218307912L;

    public RealmClientException(String msg) {
      super(msg);
    }

    public RealmClientException(String msg, Throwable cause) {
      super(msg, cause);

    }
  }

  public RealmClient() {
  }

  private List<Layer> convertRawDoc(Document doc, String url, Date lastModified)
    throws MalformedInputException {
    NodeList recs = doc.getElementsByTagName("record");
    if (recs.getLength() == 0) {
      logger.warn("Parent response contains no 'record' elements");
    }
    List<Layer> convDocs = new ArrayList<Layer>(recs.getLength());
    for (int i = 0; i < recs.getLength(); i++) {
      Element origRec = (Element) recs.item(i);
      Element layer = getFirstElemWithAttr(origRec, "layer", "name", "final");
      if (layer == null) {
        logger.warn("Ignoring record no. " + i + " from parent "
          + url + ", since no 'final' layer found.");
        continue;
      }
      Element idElem = getFirstChildByName(layer, "id");
      String recId = idElem.getTextContent().trim();
      Layer l = null;
      try {
        l = Layer.fromStoredDOM(layer);
      } catch (MalformedInputException mie) {
        throw new MalformedInputException("Malformed record '"+recId+"': "
          +mie.getMessage(), mie);
      }
      l.setId(recId);
      l.setName("original");
      l.setRealmName("world");
      l.setWorldId(null);
      try {
        l.setField("originalUri", getBaseUrl(url) + URIHelper.encode(recId) + "/");
      } catch (UnsupportedEncodingException ex) {
        throw new MalformedInputException("Cannot encode 'originalUri' field", ex);
      }
      // if the record does not contain the date, set it
      if (l.getLastModified() == null)
        l.setLastModified(lastModified);
      convDocs.add(l);
    }
    logger.debug("Retrieved " + convDocs.size() + " records for parent '" + url
      + "'");
    return convDocs;
  }

  private String getBaseUrl(String url) {
    int i = url.lastIndexOf('?');
    if (i != -1) url = url.substring(0, i);
    return url.endsWith("/") ? url : url + "/";
  }
  
  public Callable<List<Layer>> fetchAndCacheRecordsAsync(final String parentUrl, 
    final int depth, 
    final ParentRealm pr, 
    final RealmReaderWriter rw, 
    final Date lastRefreshed) 
    throws RealmClientException, InternalIOException {
    return new Callable<List<Layer>>() {
      @Override
      public List<Layer> call() throws InternalIOException {
        try {
          return RealmClient.this.fetchAndCacheRecords(parentUrl, depth, pr, rw, lastRefreshed);
        } catch (InternalIOException ioe) {
          logger.warn("Background update for parent '"+parentUrl+"' failed",
            ioe);
          throw ioe;
        }
      }
    };
  }
  
  public List<Layer> fetchAndCacheRecords(final String parentUrl, 
    int depth, 
    final ParentRealm pr, 
    final RealmReaderWriter rw, 
    final Date lastRefreshed) 
    throws RealmClientException, InternalIOException {
    Date timestap = new Date();
    Response res = this.retrieveRecords(parentUrl, depth, lastRefreshed);
    if (res != null) { //normal response, cache
      if (res.getResponseCode() != 304) {
        if (pr.getId() == null) {//we are dealing with new parent
          pr.setCreated(new Date());
          rw.writeParentRealm(pr);
        }
        rw.writeWorldRecordsForParent(res.getRecords(), pr.getId());
        return res.getRecords();
      } else { //not-modified, attempt to update cache timestamp
        Pair<List<Layer>, Date> cached = rw.readWorldRecordsForParent(pr.getId());
        if (cached == null) //impossible
          throw new InternalIOException("Missing cached records for " + pr.
            getName() + "<" + parentUrl + ">");
        cached.setSecond(timestap);
        return cached.getFirst();
      }
    } else { //nothing retrieved, possibly already in cache
      Pair<List<Layer>, Date> cached = rw.readWorldRecordsForParent(pr.getId());
      if (cached == null) //it's not
        throw new InternalIOException("Missing cached records for " + pr.
          getName() + "<" + parentUrl + ">");
      return cached.getFirst();
    }
  }
  
  //helper respnse wrapper class
  
  private static class Response {
    private List<Layer> records;
    private int responseCode;

    public Response(
      List<Layer> records, int responseCode) {
      this.records = records;
      this.responseCode = responseCode;
    }

    public List<Layer> getRecords() {
      return records;
    }

    public int getResponseCode() {
      return responseCode;
    }
    
  }

  /**
   * Fetch the world list from another TORUS
   * @return fetched world list or NULL if the world list does not need fetching
   * e.g already fetched by another thread.
   */
  private Response retrieveRecords(String baseUrl, int depth, Date lastModifiedDate)
    throws RealmClientException {
    if (depth > 0) {
      char sep = baseUrl.contains("?") ? '&' : '?';
      baseUrl += sep + "recursive=" + depth;
    }
    String ongoing = ongoingRequest.putIfAbsent(baseUrl, baseUrl);
    if (ongoing != null) { //another request in progress
      logger.info("Detected ongoing request to parent '"+baseUrl+"', standing by..");
      int waited = 0;
      while (waited < ONGOING_WAIT_TOTAL) { //timeout reached?
        waited += ONGOIGN_WAIT_CHECK;
        try {
          Thread.sleep(ONGOIGN_WAIT_CHECK);
          if (ongoingRequest.get(baseUrl) == null) {
              logger.info("Ongoing request to parent '"+baseUrl+"' has completed.");
            return null;
          }
          logger.info("Ongoing request to parent '"+baseUrl+"' still in progress...");
        } catch (InterruptedException ie) {
          throw new RealmClientException("Client thread interrupted while waiting for ongoing request to complete", ie);
        }
      }
      logger.warn("Ongoing request to parent "+baseUrl+" did not complete under timeout, cache will be inspected before returning an error.");
      return null; //we give up and signal to the caller he should check the cache anyway
    }
    HttpURLConnection conn = null;
    try {
      URL url = new URL(baseUrl);
      logger.debug("Opening connection to " + baseUrl);
      //System.setProperty("http.agent", "");
      conn = (HttpURLConnection) url.openConnection();
      conn.setRequestProperty("User-Agent", "Torus/2.0 (IndexData Aps)");
      String auth = url.getUserInfo();
      if (auth != null) { // Why can't the openConnection take care of this ??!##
        String encoded = new String(Base64.encode(auth));
        String authstr = "Basic " + encoded;
        logger.debug("Authentication '" + authstr + "' from '" + auth + "'");
        conn.setRequestProperty("Authorization", authstr);
      }
      if (lastModifiedDate != null) { //conditional-request
        try {
          String lastModified = 
            DateUtil.serialize(lastModifiedDate, DateUtil.DateTimeFormat.RFC_GMT);
          logger.info("Conditional request If-Modified-Since: "+lastModified+
            ", for '"+baseUrl+"'");
          conn.setRequestProperty("If-Modified-Since", lastModified);
        } catch (ParseException pe) {
          logger.warn("Cannot serialize lastModified date");
        }
      }
      conn.setRequestMethod("GET");
      int resCode = conn.getResponseCode();
      if (resCode == 200) {
        InputStream is = conn.getInputStream();
        Document doc = XmlUtils.parse(is);
        if (!"records".equals(doc.getDocumentElement().getNodeName()))
          throw new RealmClientException(
            "Parent response must contain 'records' as the root element");
        Date respLastMod = conn.getLastModified() > 0
          ? new Date(conn.getLastModified())
          : null;
        return new Response(convertRawDoc(doc, baseUrl, respLastMod), resCode);
      } else if (resCode == 304) {//not-modified
        try {
          logger.info("Content at '"+baseUrl+"' was not modified since '" + 
            DateUtil.serialize(lastModifiedDate, DateUtil.DateTimeFormat.RFC_GMT) +
            "', completing.");
        } catch (ParseException pe) {
          logger.warn("Cannot serialize lastModified date");
        }
        return new Response(null, 304);
      } else {
        if (resCode == 301 || resCode == 302 || resCode == 307 || resCode == 308) {
          logger.warn("Unhandled (cross-protocol?) parent redirect to "
            +conn.getHeaderField("Location"));
        }
        logger.warn("Bad parent response: " + resCode + " from " + baseUrl);
        throw new RealmClientException("Bad parent response: " + resCode
          + " from " + baseUrl);
      }
    } catch (MalformedURLException mue) {
      throw new RealmClientException("Bad parent url '" + baseUrl + "'", mue);
    } catch (IOException ioe) {
      throw new RealmClientException("Fetching data from " + baseUrl
        + " failed (" + ioe.getMessage() + ")", ioe);
    } catch (SAXException se) {
      throw new RealmClientException("Parsing response from " + baseUrl
        + " failed", se);
    } catch (MalformedInputException mre) {
        throw new RealmClientException(mre.getMessage(), mre);
    } finally {
      if (conn != null) conn.disconnect();
      ongoingRequest.remove(baseUrl);
    }
  }
}
