/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */

package com.indexdata.torus2.model;

import org.apache.log4j.Logger;
import com.indexdata.torus2.util.DateUtil;
import java.util.Date;
import com.indexdata.torus2.exception.MalformedInputException;
import com.indexdata.torus2.exception.InternalIOException;
import com.indexdata.torus2.util.DOMSerializable;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.ParseException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import static com.indexdata.torus2.util.DOMHelper.createChild;

/**
 * Represents parent realm entity.
 * @author jakub
 */
public class ParentRealm implements DOMSerializable {
  private String id;
  private String name;
  private URL url;
  private int priority = 99;
  private Date lastUpdated;
  private Date lastRefreshed;
  private int refreshAfter = 0;
  private Date created;
  public enum Trilean {TRUE, FALSE, UNDEFINED};
  private Trilean autoInherit = Trilean.UNDEFINED;
  private Logger logger = Logger.getLogger("com.indexdata.torus2");

  public ParentRealm(String id, String name, URL url) {
    this.id = id;
    this.name = name;
    this.url = url;
  }

  public ParentRealm(String id, String name, String url) throws MalformedURLException {
    this.id = id;
    this.name = name;
    this.url = new URL(url);
  }

  public String getId() {
    return id;
  }

  public void setId(String id) {
    this.id = id;
  }

  public String getName() {
    return name;
  }

  public URL getUrl() {
    return url;
  }

  public void setUrl(URL url) {
    this.url = url;
  }

  public URL resolveUrl(String host) throws InternalIOException {
    try {
      return new URL(url.toString().replaceAll("\\$\\{thisHost\\}", host));
    } catch (MalformedURLException mue) {
      throw new InternalIOException("Cannot resolve url '"+url.toString()
        +"' within host '"+host+"'", mue);
    }
  }

  public int getPriority() {
    return priority;
  }

  public void setPriority(int priority) {
    this.priority = priority;
  }

  public Date getLastUpdated() {
    return lastUpdated;
  }

  public void setLastUpdated(Date lastUpdated) {
    this.lastUpdated = lastUpdated;
  }

  public Date getLastRefreshed() {
    return lastRefreshed;
  }

  public void setLastRefreshed(Date lastRefreshed) {
    this.lastRefreshed = lastRefreshed;
  }

  public Date getCreated() {
    return created;
  }

  public void setCreated(Date created) {
    this.created = created;
  }

  public int getRefreshAfter() {
    return refreshAfter;
  }

  public void setRefreshAfter(int refreshAfter) {
    this.refreshAfter = refreshAfter;
  }
  
  public Trilean getAutoInherit() {
    return autoInherit;
  }

  public void setAutoInherit(Trilean autoInherit) {
    this.autoInherit = autoInherit;
  }

  public boolean needsRefresh(Date lR) {
    if (refreshAfter <= 0) return false;
    if (lR == null) return true;
    Date now = new Date();
    long rA = refreshAfter * 1000;
    Date nextRefresh = new Date(lR.getTime()+rA);
    logger.debug("Parent '"+name+"': nextRefresh='"+nextRefresh+"', now='"+now+"'");
    return nextRefresh.before(now);
  }
  
  @Override
  public Element toDOM(Element parentElement) {
    Element parentNode = createChild(parentElement, "parent");
    parentNode.setAttribute("id", id);
    parentNode.setAttribute("name", name);
    parentNode.setAttribute("url", url.toString());
    parentNode.setAttribute("priority", Integer.toString(priority));
    if (lastUpdated != null)
      parentNode.setAttribute("lastUpdated", DateUtil.serialize(lastUpdated));
    //lastRefreshed is only set on responses
    if (lastRefreshed != null)
      parentNode.setAttribute("lastRefreshed", DateUtil.serialize(lastRefreshed));
    parentNode.setAttribute("refreshAfter", Integer.toString(refreshAfter));
    if (created != null)
      parentNode.setAttribute("created", DateUtil.serialize(created));
    if (autoInherit.equals(Trilean.TRUE))
      parentNode.setAttribute("autoInherit", "yes");
    if (autoInherit.equals(Trilean.FALSE))
      parentNode.setAttribute("autoInherit", "no");
    return parentNode;
  }

  /**
   * This method can be safely called on user input as it ignores system
   * controlled fields that end-up serialized in the output document.
   * @param doc
   * @param id
   * @return
   * @throws MalformedInputException
   */
  public static ParentRealm fromDOM(Document doc, String id) throws MalformedInputException {
    Element root = doc.getDocumentElement();
    if (!"parent".equals(root.getTagName()))
      throw new MalformedInputException("Parent realm document must contain 'parent' root element");
   String name = root.getAttribute("name");
   if (name == null || name.isEmpty())
     throw new MalformedInputException("Missing attribute 'name'");
   String url = root.getAttribute("url");
   if (url == null || url.isEmpty())
     throw new MalformedInputException("Missing attribute 'url'");
   URL parsedUrl = null;
   try {
    parsedUrl = new URL(url);
   } catch (MalformedURLException mue) {
     throw new MalformedInputException("Malformed attribute 'url'", mue);
   }
   ParentRealm pr = new ParentRealm(id, name, parsedUrl);
   String prio = root.getAttribute("priority");
   if (prio != null && !prio.isEmpty()) {
     try {
       pr.setPriority(Integer.parseInt(prio));
     } catch (NumberFormatException nfe) {
       throw new MalformedInputException("'priority' attribute must be an int", nfe);
     }
   }
   String refreshAfter = root.getAttribute("refreshAfter");
   if (refreshAfter != null && !refreshAfter.isEmpty()) {
     try {
      int rA = Integer.parseInt(refreshAfter);
      pr.setRefreshAfter(rA);
     } catch (NumberFormatException nfe) {
       throw new MalformedInputException("'refreshAfter' must be an int", nfe);
     }
   }
   String autoInheritStr = root.getAttribute("autoInherit");
   if (autoInheritStr != null && autoInheritStr.equals("yes"))
     pr.setAutoInherit(Trilean.TRUE);
   if (autoInheritStr != null && autoInheritStr.equals("no"))
     pr.setAutoInherit(Trilean.FALSE);
   //otherwise inherit realm flag
   return pr;
  }
  
  /**
   * This method calls ParentRealm#fromDOM and also parses the stored fields
   * that otherwise are ignored from the user request.
   * @param doc input DOM doc
   * @param id of the entity
   * @return new instance of ParentRealm
   * @throws MalformedInputException
   */
  public static ParentRealm fromStoredDOM(Document doc, String id) 
    throws MalformedInputException {
    ParentRealm pr = ParentRealm.fromDOM(doc, id);
    String lr = doc.getDocumentElement().getAttribute("lastUpdated");
    if (lr != null && !lr.isEmpty()) {
      try {
        Date lastUpdated = DateUtil.parse(lr);
        pr.setLastUpdated(lastUpdated);
      } catch (ParseException pe) {
        throw new MalformedInputException("Error parsing 'lastUpdated' date", pe);
      }
    }
    String cr = doc.getDocumentElement().getAttribute("created");
    if (cr != null && !cr.isEmpty()) {
      try {
        Date created = DateUtil.parse(cr);
        pr.setCreated(created);
      } catch (ParseException pe) {
        throw new MalformedInputException("Error parsing 'created' date", pe);
      }
    }
    return pr;
  }

}
