/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */
package com.indexdata.torus2.model;

import com.indexdata.torus2.util.DateUtil;
import java.text.ParseException;
import java.util.Date;
import java.util.Set;
import com.indexdata.torus2.exception.MalformedInputException;
import com.indexdata.torus2.util.DOMSerializable;
import com.indexdata.torus2.util.XMLSerializable;
import com.indexdata.torus2.util.XMLWriter;
import java.util.ListIterator;
import com.indexdata.utils.XmlUtils;
import java.util.HashSet;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.w3c.dom.Element;

import static com.indexdata.torus2.util.DOMHelper.getChildrenByName;
import static com.indexdata.torus2.util.DOMHelper.createChild;

/**
 *
 * @author jakub
 */
public class Record implements DOMSerializable, XMLSerializable {
  private String type;
  private String parentId;
  private List<Layer> layers = new LinkedList<Layer>();
  private Set<String> hiddenLayers = new HashSet<String>();
  private Date lastModified;

  public Record(String type, String parentId) {
    this.type = type;
    this.parentId = parentId;
  }

  public Record(String type, String parentId, Layer layer) {
    this.type = type;
    this.parentId = parentId;
    layers.add(layer);
  }

  public String getType() {
    return type;
  }

  public String getParentId() {
    return parentId;
  }

  public void setLastModified(Date lastModified) {
    this.lastModified = lastModified;
  }

  public Date getLastModified() {
    return lastModified;
  }

  public boolean hideLayer(String layerName) {
    return hiddenLayers.add(layerName);
  }

  public boolean showLayer(String layerName) {
    return hiddenLayers.remove(layerName);
  }

  public Layer getLayer(String layerName) {
    for (Layer l : layers) {
      if (l.name.equals(layerName)) {
        return l;
      }
    }
    return null;
  }

  public Record setLayer(Layer layer) {
    ListIterator<Layer> it = layers.listIterator();
    while (it.hasNext()) {
      if (it.next().getName().equals(layer.name)) {
        it.remove();
        it.add(layer);
        return this;
      }
    }
    layers.add(layer);
    return this;
  }

  public Layer addLayer(String name) {
    Layer l = new Layer(name);
    layers.add(l);
    return l;
  }
  
  public Layer addLayer(Layer layer) {
    layers.add(layer);
    return layer;
  }

  public List<Layer> getLayers() {
    return layers;
  }

  public static class Layer implements Iterable<Layer.Field> {
    private String name;
    private String id;
    private String realmName;
    private String worldId;
    private Date lastModified;
    private Date creationDate;
    private boolean disabled;
    //only valid for master record layers
    private String matchQuery;
    private String label;

    //we use a list to retain user order
    //TODO consider changing this to TreeMap
    private List<Field> fields = new LinkedList<Field>();

    public Layer(String name) {
      this.name = name;
    }

    public String getId() {
      return id;
    }

    public void setId(String id) {
      this.id = id;
    }

    public String getName() {
      return name;
    }

    public void setName(String name) {
      this.name = name;
    }

    public String getRealmName() {
      return realmName;
    }

    public void setRealmName(String realmName) {
      this.realmName = realmName;
    }

    public String getWorldId() {
      return worldId;
    }

    public void setWorldId(String worldId) {
      this.worldId = worldId;
    }

    public Date getLastModified() {
      return lastModified;
    }

    public void setLastModified(Date lastModifed) {
      this.lastModified = lastModifed;
    }
    
    public Date getCreationDate() {
      return creationDate;
    }

    public void setCreationDate(Date creationDate) {
      this.creationDate = creationDate;
    }

    public String getMatchQuery() {
      return matchQuery;
    }

    public void setMatchQuery(String matchQuery) {
      this.matchQuery = matchQuery;
    }

    public void setLabel(String label) {
      this.label = label;
    }

    public String getLabel() {
      return label;
    }

    public boolean isDisabled() {
      return disabled;
    }

    public void setDisabled(boolean disabled) {
      this.disabled = disabled;
    }
    
    public Field getField(String name) {
      //search for existing
      for (Field f : fields) {
        if (f.name.equals(name)) {
          return f;
        }
      }
      //include psudo-fields in searching
      if ("id".equals(name) && id != null) return new Field("id", id);
      if ("worldId".equals(name) && worldId != null)
        return new Field("worldId", worldId);
      if ("realm".equals(name) && realmName != null)
        return new Field("realm", realmName);
      if ("lastModified".equals(name) && lastModified != null)
        return new Field("lastModified", DateUtil.serialize(lastModified));
      if ("creationDate".equals(name) && creationDate != null)
        return new Field("creationDate", DateUtil.serialize(creationDate));
      if ("disabled".equals(name) && disabled)
        return new Field("disabled", "yes");
      return null;
    }

    @Override
    public Iterator<Field> iterator() {
      return new Iterator<Field>() {
        private int i;

        @Override
        public boolean hasNext() {
          return i < fields.size();
        }

        @Override
        public Field next() {
          return fields.get(i++);
        }

        @Override
        public void remove() {
          fields.remove(i);
        }
      };
    }

    public Field setField(String name, String contents) {
      //search for existing
      for (Field f : fields) {
        if (f.name.equals(name)) {
          f.contents = contents;
          return f;
        }
      }
      Field f = new Field(name, contents);
      fields.add(f);
      return f;
    }

    public Field setGoneField(String name) {
      //search for existing
      for (Field f : fields) {
        if (f.name.equals(name)) {
          f.gone = true;
          return f;
        }
      }
      Field f = new Field(name);
      fields.add(f);
      return f;
    }

    public void toXML(XMLWriter xw, String type) {
      xw
        .tag("layer").attr("name", name)
          .attr("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
          .attr("xsi:type", type+"TypeLayer")
          .tag("id").text(id).end()
          .tag("realm").text(realmName).end();
      if (worldId != null) xw.tag("worldId").text(worldId).end();
      for (Layer.Field f : fields) {
        if ("id".equals(f.name) || "realm".equals(f.name)
          || "worldId".equals(f.name)) continue;
        xw.tag(f.name).text(f.contents).end();
      }
      xw.end();
    }

    public Element toDOM(Element parent, String type, String parentId) {
      Element lE = createChild(parent, "layer");
      lE.setAttribute("name", name);
      lE.setAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance");
      lE.setAttribute("xsi:type", type+"TypeLayer");
      //TODO introduce layer polymorphism
      if ("master-override".equals(name)) {
        //null values are rendered as empty to protect clients
        lE.setAttribute("matchQuery", matchQuery != null ? matchQuery : "");
        lE.setAttribute("label", label != null ? label : "");
      }
      String id = !name.equals("original") || parentId == null || parentId.isEmpty()
        ? this.id : parentId + "." + this.id;
      XmlUtils.appendTextNode(lE, "id", id);
      if (realmName != null)
        XmlUtils.appendTextNode(lE, "realm", realmName);
      if (worldId != null)
        XmlUtils.appendTextNode(lE, "worldId", worldId);
      if (lastModified != null)
        XmlUtils.appendTextNode(lE, "lastModified",
          DateUtil.serialize(lastModified));
      if (creationDate != null)
        XmlUtils.appendTextNode(lE, "creationDate",
          DateUtil.serialize(creationDate));
      if (disabled)
        XmlUtils.appendTextNode(lE, "disabled", "yes");
      for (Layer.Field f : fields) {
        //the Entity id/realm/worldId takes precedence
        if ("id".equals(f.name) || "realm".equals(f.name)
          || "worldId".equals(f.name) || "lastModified".equals(f.name) 
          || "creationDate".equals(f.name) || "disabled".equals(f.name))
          continue;
        XmlUtils.appendTextNode(lE, f.name, f.contents);
      }
      return lE;
    }
    
    public static Layer fromDOM(Element layerNode) throws MalformedInputException {
      return fromDOM(layerNode, false);
    }
    
    public static Layer fromStoredDOM(Element layerNode) throws MalformedInputException {
      return fromDOM(layerNode, true);
    }

    private static Layer fromDOM(Element layerNode, boolean isStored) throws MalformedInputException {
      String layerName = layerNode.getAttribute("name");
      if (layerName == null || layerName.isEmpty()) {
        throw new MalformedInputException(
          "element 'layer' missing non-empty 'name' attribute");
      }
      Layer l = new Layer(layerName);
      if (layerName.equals("master-override")) {
        //null values are used to express lack of update
        if (layerNode.hasAttribute("matchQuery"))
          l.setMatchQuery(layerNode.getAttribute("matchQuery"));
        if (layerNode.hasAttribute("label"))
          l.setLabel(layerNode.getAttribute("label"));
      }
      NodeList children = layerNode.getChildNodes();
      for (int i = 0; i < children.getLength(); i++) {
        Node child = children.item(i);
        if (child.getNodeType() == Node.ELEMENT_NODE) {
          Element elem = (Element) child;
          //ignore id/realm in DOM!
          if ("id".equals(elem.getTagName()) || "realm".equals(elem.getTagName())) {
            continue;
          }
          //use worldId
          if ("worldId".equals(elem.getTagName())) {
            String val = elem.getTextContent().trim();
            if (val != null && !val.isEmpty())
              l.worldId = val;
            continue;
          }
          // pick up the last modified
          if ("lastModified".equals(elem.getTagName())) {
            if (!isStored) continue; //user updates not allowed
            Date lM;
            try {
              lM = DateUtil.parse(elem.getTextContent());
            } catch (ParseException ex) {
              throw new MalformedInputException("Malformed 'lastModified' field", ex);
            }
            l.setLastModified(lM);
            continue;
          }
          // pick up creation
          if ("creationDate".equals(elem.getTagName())) {
            if (!isStored) continue; //user updates not allowed
            Date cD;
            try {
              cD = DateUtil.parse(elem.getTextContent());
            } catch (ParseException ex) {
              throw new MalformedInputException("Malformed 'creationDate' field", ex);
            }
            l.setCreationDate(cD);
            continue;
          }
          //pickup disabled
          if ("disabled".equals(elem.getTagName())) {
            boolean disabled = elem.getTextContent().equals("yes");
            l.setDisabled(disabled);
            continue;
          }
          //check for removed elems
          String gone = elem.getAttribute("gone");
          if (gone != null && gone.equals("yes")) {
            l.setGoneField(elem.getTagName());
          } else {
            l.setField(elem.getTagName(), elem.getTextContent().trim());
          }
        }
      }
      return l;
    }

    public class Field {
      private String name;
      private String contents;
      private boolean gone = false;

      private Field(String name, String contents) {
        this.name = name;
        this.contents = contents;
      }

      private Field(String goneFieldName) {
        this.name = goneFieldName;
        this.gone = true;
      }

      public String getName() {
        return name;
      }

      public String getContents() {
        return contents;
      }

      public void setContents(String contents) {
        this.contents = contents;
      }

      public boolean isGone() {
        return gone;
      }
    }
  }

  @Override
  public void toXML(XMLWriter xw) {
    xw
      .tag("record").attr("type", type);
    for (Layer l : layers) {
      l.toXML(xw, type);
    }
    xw.end();
  }

  @Override
  public Element toDOM(Element parent) {
    Element recordElem = createChild(parent, "record");
    recordElem.setAttribute("type", type);
    for (Layer l : layers) {
      if (!hiddenLayers.contains(l.getName()))
        l.toDOM(recordElem, type, parentId);
    }
    return recordElem;
  }

  public static Record fromDOM(Element root, String type) throws
    MalformedInputException {
    if (!"record".equals(root.getTagName())) {
      throw new MalformedInputException(
        "Record must contain 'record' as root element");
    }
    List<Element> layerNodes = getChildrenByName(root, "layer");
    if (layerNodes.isEmpty()) {
      throw new MalformedInputException(
        "element 'record' contains no 'layer' children elements");
    }
    //set the parentId to null initially
    Record r = new Record(type, null);
    for (Element lN : layerNodes) {
     Layer l = Layer.fromDOM(lN);
     r.addLayer(l);
    }
    return r;
  }
}
