package com.indexdata.torus2.query;

import com.indexdata.torus2.model.Record;
import com.indexdata.torus2.model.Record.Layer;
import com.indexdata.torus2.model.Record.Layer.Field;
import com.indexdata.torus2.util.DateUtil;
import java.text.Collator;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.log4j.Logger;

import org.z3950.zing.cql.CQLSortNode;
import org.z3950.zing.cql.Modifier;
import org.z3950.zing.cql.ModifierSet;

/**
 * Comparator, based on criteria from provided CQLSortNode.
 * 
 * Compares XML documents and is used for sorting XML documents in 
 * ascending or descending order by given Elements in the documents.
 * 
 * Comparison is based on the text content of the Elements
 * 
 * If the field to sort by is not found in one of the documents 
 * under comparison, the comparison will be done using the empty string. 
 * If there are multiple elements with the same name, the first 
 * occurrence will be used for comparison. 
 * 
 * The comparator supports multiple sort fields but only the modifiers
 * sort order (default ascending) and locale (default is the systems default).
 * 
 * Sorting will be character based (numeric sorting not supported).
 * 
 * @author Niels Erik
 */
public class RecordComparator implements Comparator<Record> {
  private static String[] supportedModifiersArray = {"sort.descending",
    "sort.ascending",
    "sort.locale",
    "date"};
  List<ModifierSet> modifierSets = null;
  HashMap<ModifierSet, HashMap<String, String>> mapOfModifierMaps =
    new HashMap<ModifierSet, HashMap<String, String>>();
  private static Logger logger = Logger.getLogger("com.indexdata.torus2");
  private final String[] layerNames;

  public RecordComparator(CQLSortNode cqlSortNode, String... layerNames) {
    this.modifierSets = cqlSortNode.getSortIndexes();
    ArrayList<String> supportedModifiers = new ArrayList<String>(Arrays.asList(
      supportedModifiersArray));
    // Prepare a key/value map holding the modifiers for each sort field(also called ModifierSet),
    // Put this map in another map keyed by the sort field
    // All of this to minimize repetitive computation of modifiers for each pair
    // of documents that are compared during sorting
    for (ModifierSet modifierSet : this.modifierSets) {
      List<Modifier> modifiers = modifierSet.getModifiers();
      HashMap<String, String> modifierMap = new HashMap<String, String>();
      for (Modifier modifier : modifiers) {
        if (supportedModifiers.contains(modifier.getType())) {
          modifierMap.put(modifier.getType().toLowerCase(), modifier.getValue());
        } else {
          logger.warn("Unsupported modifier: " + modifier.getType());
        }
      }
      mapOfModifierMaps.put(modifierSet, modifierMap);
    }
    this.layerNames = layerNames;
  }

  /**
   * Will compare two documents based on alphabetical comparison of the content of the specified fields
   * using the given sort order and locale if any.
   *
   */
  @Override
  public int compare(Record rec1, Record rec2) {
    int result = 0;
    // Loop the sort fields
    for (ModifierSet modifierSet : this.modifierSets) {
      String fieldName = modifierSet.getBase();
      //sort on a field from the specified layer, if none specified fallback to the first
      Layer layerA = null;
      if (rec1.getLayers().size() > 0) {
        for (int i=0; i<layerNames.length; i++) {
          layerA = rec1.getLayer(layerNames[i]);
          if (layerA != null) break;
        }
        if (layerA == null) {
          layerA = rec1.getLayers().get(0);
        }
      }
      Layer layerB = null;
      if (rec2.getLayers().size() > 0) {
        for (int i=0; i<layerNames.length; i++) {
          layerB = rec2.getLayer(layerNames[i]);
          if (layerB != null) break;
        }
        if (layerB == null) {
          layerB = rec2.getLayers().get(0);
        }
      }
      Field fieldA = layerA != null
        ? layerA.getField(fieldName) : null;
      Field fieldB = layerB != null
        ? layerB.getField(fieldName) : null;
      //empty string will put records without the sort field in front, is that we want?
      String field1 = fieldA != null ? fieldA.getContents() : "";
      String field2 = fieldB != null ? fieldB.getContents() : "";
      logger.debug("sortby '"+fieldName+"' comparing [" + field1 + "] to [" + field2 + "]");
      // sort modifiers for this field
      Map<String, String> modifierMap = mapOfModifierMaps.get(modifierSet);
      int sortDirection = (modifierMap.containsKey("sort.descending") ? -1 : 1);
      //date sorting
      if (modifierMap.containsKey("date")) {
        //yes, this will be parsing it log(n) times
        //TODO try to implement smart date parsing caching
        Date d1, d2;
        try {
          d1 = DateUtil.parse(field1);
        } catch (ParseException pe) {
          logger.warn("Cannot parse date from '"+field1+"', assume empty.");
          return sortDirection;
        }
        try {
          d2 = DateUtil.parse(field2);
        } catch (ParseException pe) {
          logger.warn("Cannot parse date from '"+field2+"', assume empty.");
          return -sortDirection;
        }
        result = sortDirection * d1.compareTo(d2);
      } else {
        //character-based sorting
        Locale locale = (modifierMap.containsKey("sort.locale") 
          ? new Locale(modifierMap.get("sort.locale").toString()) 
          : null);
        Collator collator = (locale == null) ? Collator.getInstance() 
          : Collator.getInstance(locale);
        // collate
        result = sortDirection * collator.compare(field1, field2);
      }
      if (result != 0) {
        // We're done, unless the two documents are equal on this field
        break;
      }
    }
    return result;
  }
}
