/*
 * Copyright (c) 1995-2012, Index Datassss
 * All rights reserved.
 * See the file LICENSE for details.
 */
package com.indexdata.torus2.query.match.term;

import com.indexdata.torus2.query.QueryException;
import com.indexdata.torus2.query.match.TermMatcher;
import com.indexdata.torus2.util.DateUtil;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.log4j.Logger;


/**
 *
 * @author jakub
 */
public class DateWithinMatcher implements TermMatcher {
  private Date f;
  private Date t;
  private DateUtil.DateTimeFormat dtf;
  private static Logger logger = Logger.getLogger("com.indexdata.torus2");
  
  public DateWithinMatcher(String term, String dateFormat) throws QueryException {
    String from = null, to = null;
            String[] dates = term.split("\\.\\.");
            if (dates.length == 0 || dates.length > 2) 
              throw new QueryException("CQL error: within/rfcDate requires "
                + "a date string or two date strings seperated wtih '..', given '"+term+"'");
            if (dates.length == 1)
              from = dates[0];
            if (dates.length == 2) {
              from = dates[0];
              to = dates[1];
            }
    if ("rfcDate".equalsIgnoreCase(dateFormat))
      dtf = DateUtil.DateTimeFormat.RFC;
    else if ("isoDate".equalsIgnoreCase(dateFormat))
      dtf = DateUtil.DateTimeFormat.ISO;
    else
      throw new QueryException("CQL error: unsupported date format '"
          +dateFormat+"'");
   try {
      if (from != null && !from.isEmpty()) {
        //check if the upper bound is a timeslice and if not, use as baseline
        if (to != null && !to.isEmpty() && !isTimeSlice(to)) {
          t = DateUtil.parse(to, dtf);
        }
        f = isTimeSlice(from) ? parseFromTimeSlice(from, t) : DateUtil.parse(from, dtf);
      }
      //unless already parsed
      if (t == null && to != null && !to.isEmpty()) {
        t = isTimeSlice(to) ? parseFromTimeSlice(to, isTimeSlice(from) ? null : f) : DateUtil.parse(to, dtf);
      }
    } catch (ParseException pe) {
      throw new QueryException("Parsing 'from' or 'to' date failed,"
        + " check if the query syntax matches date format.", pe);
    }
  }
  
  @Override
  public boolean match(String value) {
    Date d = null;
    try {
      d = DateUtil.parse(value);
      return ((f == null || f.equals(d) || f.before(d)) 
           && (t == null || t.equals(d) || t.after(d)));
    } catch (ParseException ex) {
      logger.warn("Parsing date from field '"+value+"' failed, ignoring.");
      return false;
    }
  }
  
  private final static Pattern beforeAfter = Pattern
    .compile("^[+-](?:\\s*(\\d+)y)?(?:\\s*(\\d+)M)?(?:\\s*(\\d+)d?)?(?:\\s*(\\d+)h)?(?:\\s*(\\d+)m)?(?:\\s*(\\d+)s)?$");
  //1 - y, 2 - M, 3 - d, 4 - h, 5 - m, 6 - s
  
  private boolean isTimeSlice(String dateStr) {
    return dateStr.startsWith("+") || dateStr.startsWith("-");
  }

  private Date parseFromTimeSlice(String dateStr, Date baseline) throws ParseException {
    Matcher m = beforeAfter.matcher(dateStr);
    if (m.matches()) {
      int sign = dateStr.startsWith("-") ? -1 : 1;
      Calendar c = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
      if (baseline != null) {
        c.setTimeInMillis(baseline.getTime());
      }
      //year
      int yi = 0;
      String ys = m.group(1);
      if (ys != null) {
        yi = Integer.parseInt(ys);
      }
      c.set(Calendar.YEAR, c.get(Calendar.YEAR)+yi*sign);
      //month
      int moi = 0;
      String mos = m.group(2);
      if (mos != null) {
        moi = Integer.parseInt(mos);
      }
      c.set(Calendar.MONTH, c.get(Calendar.MONTH)+moi*sign);
      //day
      int di = 0;
      String ds = m.group(3);
      if (ds != null) {
        di = Integer.parseInt(ds);
      }
      c.set(Calendar.DAY_OF_YEAR, c.get(Calendar.DAY_OF_YEAR)+di*sign);
      //hour
      int hi = 0;
      String hs = m.group(1);
      if (hs != null) {
        hi = Integer.parseInt(hs);
      }
      c.set(Calendar.HOUR_OF_DAY, c.get(Calendar.HOUR_OF_DAY)+hi*sign);
      //mins
      int mi = 0;
      String ms = m.group(1);
      if (ms != null) {
        mi = Integer.parseInt(ms);
      }
      c.set(Calendar.MINUTE, c.get(Calendar.MINUTE)+mi*sign);
      //secs
      int si = 0;
      String ss = m.group(1);
      if (ss != null) {
        si = Integer.parseInt(ss);
      }
      c.set(Calendar.SECOND, c.get(Calendar.SECOND)+si*sign);
      if (logger.isDebugEnabled()) {
        logger.debug("Time slice boundary calculated to "+DateUtil.serialize(c.getTime()));
      }
      return c.getTime();
    }
    throw new ParseException("Bad time slice '"+dateStr+"'", 0);
  }
 
}
