/*
 * Copyright (c) 1995-2011, Index Data
 * All rights reserved.
 * See the file LICENSE for details.
 */
package com.indexdata.torus2.rw.impl;

import java.util.concurrent.ConcurrentHashMap;
import java.io.Writer;
import java.io.BufferedWriter;
import com.indexdata.torus2.exception.MalformedInputException;
import com.indexdata.torus2.model.Record.Layer;
import com.indexdata.torus2.exception.NotFoundException;
import com.indexdata.torus2.model.ParentRealm;
import com.indexdata.torus2.model.RealmDefinition;
import com.indexdata.torus2.exception.InternalIOException;
import com.indexdata.torus2.rw.RealmReaderWriter;
import com.indexdata.torus2.util.NaturalOrderComparator;
import org.apache.log4j.Logger;
import com.indexdata.utils.XmlUtils;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import javax.xml.transform.TransformerException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import static com.indexdata.torus2.util.DOMHelper.*;
import com.indexdata.torus2.util.Pair;
import java.util.Date;

/**
 * Flat, multi-file implementation of the record store
 * @author jakub
 */
public class RealmFileRW implements RealmReaderWriter {
  private final static String META_FILE = ".torus-meta";
  private String basePath;
  private String realmPath;
  private String realmName;
  private RealmDefinition rdef;
  private Logger logger = Logger.getLogger("com.indexdata.torus2");
  private final static ConcurrentHashMap<String, Object> resourceLock = 
    new ConcurrentHashMap<String, Object>();
 
  //this thread-safely initializes the lock in the map
  public Object getLock(String host) {
    Object lock = resourceLock.get(host);
    if (lock == null) {
      Object newLock = new Object();
      lock = resourceLock.putIfAbsent(host, newLock);
      if (lock == null) {
        // put succeeded, use new value
        lock = newLock;
      }
    }
    return lock;
  }
  
  /**
   * Opens RW on a specified realm. It's possible to set 'realm' argument
   * to null in which case the RW will remain closed but will allow certain
   * operation (e.g listRealms)
   * @param realmName
   * @param basePath
   */
  public RealmFileRW(String realmName, String basePath) {
    if (realmName != null) {
      //the rw will remain closed but will allow certain operations
      this.realmPath = basePath + File.separator + realmName;
      this.realmName = realmName;
    }
    this.basePath = basePath;
  }

  @Override
  public boolean create(RealmDefinition rd) throws InternalIOException {
    if (realmPath == null) return false;
    synchronized (getLock(realmPath)) {
      File realmDir = new File(realmPath);
      if (realmDir.exists()) return false;
      File recDir = new File(realmPath + File.separator + "records");
      if (!recDir.mkdirs())
        throw new InternalIOException("Cannot create 'records' dir at '"+recDir.getAbsolutePath()+"'");
      File recMeta = new File(realmPath + File.separator + "records" 
        + File.separator + META_FILE);
      try {
        storeIndex(recMeta, 0);
      } catch (IOException ex) {
        throw new InternalIOException(ex);
      }
      File parentsDir = new File(realmPath + File.separator + "parents");
      if (!parentsDir.mkdirs())
        throw new InternalIOException("Cannot create 'parents' dir at '"+parentsDir.getAbsolutePath()+"'");
      File parentMeta = new File(realmPath + File.separator + "parents" 
        + File.separator + META_FILE);
      try {
        storeIndex(parentMeta, 0);
      } catch (IOException ex) {
        throw new InternalIOException(ex);
      }
      writeDefinition(rd);
      return true;
    }
  }

  @Override
  public boolean update(RealmDefinition rd) throws InternalIOException {
    if (realmPath == null) return false;
    synchronized (getLock(realmPath)) {
      writeDefinition(rd);
    }
    return true;
  }
  
  private void writeDefinition(RealmDefinition rd) throws InternalIOException {
    //store realm record
    File realmDef = new File(realmPath + File.separator + "Realm-Definition");
    Document doc = rd.toDOM(null).getOwnerDocument();
    FileWriter fw = null;
    try {
      fw = new FileWriter(realmDef);
      XmlUtils.serialize(doc, fw);
    } catch (IOException io) {
      throw new InternalIOException("Cannot store realm definition at '"+realmDef.getAbsolutePath()+"'", io);
    } catch (TransformerException te) {
      throw new InternalIOException("Cannot serialize realm definition at '"+realmDef.getAbsolutePath()+"'", te);
    } finally {
      try {
        if (fw != null) fw.close();
      } catch (IOException io) {
        throw new InternalIOException("Flushing file failed at '"+realmDef.getAbsolutePath()+"'", io);
      }
    }
  }

  @Override
  public synchronized boolean open() throws InternalIOException {
    //to list all realms we create a 'dummy' realm with no name and path
    //this ensures open won't break (but most functionlity will be disabled)
    if (realmPath == null) return false;
    File rdF = new File(realmPath + File.separator + "Realm-Definition");
    if (!rdF.isFile())
      throw new NotFoundException("Realm-Definition does not exist or not a file at " + rdF.getAbsolutePath());
    try {
      Document rdD = XmlUtils.parse(rdF);
      rdef = RealmDefinition.fromDOM(rdD, realmName);
    } catch (SAXException ex) {
      throw new InternalIOException("Cannot parse Realm-Definition at '"+rdF.getAbsolutePath()+"'", ex);
    } catch (IOException ex) {
      throw new InternalIOException("Cannot read Realm-Definition at '"+rdF.getAbsolutePath()+"'", ex);
    } catch (MalformedInputException mie) {
      throw new InternalIOException(mie);
    }
    return true;
  }

  @Override
  public RealmDefinition describe() {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    return rdef;
  }

  private void rmdir(File f) throws IOException {
    if (f.isDirectory()) {
      for (File c : f.listFiles()) {
        rmdir(c);
      }
    }
    if (!f.delete()) {
      throw new IOException("Failed to delete file or dir at '"+f.getAbsolutePath()+"'");
    }
  }

  @Override
  public void destroy() throws InternalIOException {
    if (realmPath == null)
      throw new InternalIOException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    synchronized (getLock(realmPath)) {
      File realmDir = new File(realmPath);
      try {
        if (!realmDir.exists())
        throw new NotFoundException("Realm dir does not exist at '"+realmDir.getAbsolutePath()+"'");
        rmdir(realmDir);
      } catch (IOException ioe) {
        throw new InternalIOException("Deleting file/dir failed at '"+realmDir.getAbsolutePath()+"'", ioe);
      }
    }
  }

  private String escapeId(String id) {
    return id.replace("/", "\u2044").replace("\\", "\u2215");
  }

  private String unescapeId(String fileId) {
    return fileId.replace("\u2044", "/").replace("\u2215", "\\");
  }

  @Override
  public List<Layer> readOverrideRecords()
    throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File recsDir = new File(realmPath + File.separator + "records");
    if (recsDir.isDirectory()) {
      File[] files = recsDir.listFiles();
      List<Layer> recList = new ArrayList<Layer>(files.length);
      for (File ff : files) {
        if (ff.isFile() && !ff.getName().startsWith(".")) {
          Layer layer;
          try {
            Document doc = XmlUtils.parse(ff);
            setRecordType(doc.getDocumentElement(), rdef);
            //enforce ids/type/layerType in case data comes from migration
            //or is not fully valid
            Element overL = getFirstElemWithAttr(doc.getDocumentElement(), "layer", "name", "override");
            if (overL == null)
              throw new MalformedInputException("Record '"
                +unescapeId(ff.getName())+"'is missing 'override' layer at '"
                +ff.getAbsolutePath()+"'");
            layer = Layer.fromStoredDOM(overL);
            layer.setId(unescapeId(ff.getName()));
            layer.setRealmName(rdef.getRealmName());
          } catch (SAXException ex) {
            throw new InternalIOException("Cannot parse record contens at '"
              +ff.getAbsolutePath() +"'", ex);
          } catch (IOException ex) {
            throw new InternalIOException("Cannot read record from file at '"
              +ff.getAbsolutePath()+"'", ex);
          } catch (MalformedInputException mie) {
            throw new InternalIOException(mie);
          }
          recList.add(layer);
        }
      }
      return recList;
    } else {
      throw new NotFoundException("Missing realm directory at '" + recsDir.
        getAbsolutePath() + "'");
    }
  }

  @Override
  public Layer readOverrideRecord(String id) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File recFile = new File(realmPath + File.separator + "records"
      + File.separator + escapeId(id));
    if (!recFile.isFile()) {
      throw new NotFoundException("Cannot find file for record '" + id + "' at '"
        + recFile.getAbsolutePath() +"'");
    }
    try {
      Document doc = XmlUtils.parse(recFile);
      setRecordType(doc.getDocumentElement(), rdef);
      Element layerElem = getFirstElemWithAttr(doc.getDocumentElement(), "layer", "name", "override");
      if (layerElem == null) {
        throw new MalformedInputException("Record '"+id+"' is missing an 'override' layer at '"
          +recFile.getAbsolutePath()+"'");
      }
      Layer layer = Layer.fromStoredDOM(layerElem);
      layer.setId(id);
      layer.setRealmName(rdef.getRealmName());
      return layer;
    } catch (SAXException ex) {
      throw new InternalIOException("Cannot parse record contens at '"
        + recFile.getAbsolutePath() + "'", ex);
    } catch (IOException ex) {
      throw new InternalIOException("Cannot read record from file at '"
        + recFile.getAbsolutePath() + "'", ex);
     } catch (MalformedInputException mie) {
      throw new InternalIOException(mie);
    }
  }

  @Override
  public Layer writeOverrideRecord(String id, Layer update) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File out = new File(realmPath + File.separator + "records" +
      File.separator + escapeId(id));
    if (!out.exists())
      throw new NotFoundException("No file for record '"+out.getAbsolutePath()+"'");
    FileWriter fw = null;
    try {
      fw = new FileWriter(out);
      //we will retain the id's so the on-disk records are not confusing
      update.setId(id);
      update.setRealmName(rdef.getRealmName());
      Document envelope = XmlUtils.newDoc("record");
      setRecordType(envelope.getDocumentElement(), rdef);
      update.toDOM(envelope.getDocumentElement(), rdef.getRealmType(), null);
      XmlUtils.serialize(envelope, fw);
    } catch (IOException ioe) {
      throw new InternalIOException("Cannot write record file at '"
        + out.getAbsolutePath() +"'", ioe);
    } catch (TransformerException te) {
      throw new InternalIOException("Record contents serialization failed at '"
        + out.getAbsolutePath() +"'", te);
    } finally {
      try {
        if (fw != null) fw.close();
      } catch (IOException io) {
        throw new InternalIOException("Flushing file failed at '"
          +out.getAbsolutePath()+"'", io);
      }
    }
    return update;
  }
  
  private int readIndex(File meta) throws IOException {
    int i = 0;
    if (meta.isFile()) {
      BufferedReader reader = null;
      try {
        reader = new BufferedReader(new FileReader(meta));
        String idx = reader.readLine();
        i = Integer.parseInt(idx);
      } catch (IOException ioe) {
        logger.debug(ioe);
        logger.warn("Reading meta file at '"+meta.getAbsolutePath()+"' failed,"
          + " limping along without it..");
      } catch (NumberFormatException nfe) {
        logger.debug(nfe);
        logger.warn("Contents of meta file malformed at '"
          +meta.getAbsolutePath()+", limping along without it..");
      } finally {
        try {
          if (reader != null) reader.close();
        } catch (IOException ioe) {
          throw new IOException("Flushing meta file failed at '"
            +meta.getAbsolutePath()+"'", ioe);
        }
      }
    } else {
      logger.warn("Meta file does not exist or not a file at '"
        +meta.getAbsolutePath()+"', limping along without it..");
    }
    return i;
  }
  
  private void storeIndex(File meta, int i) throws IOException {
    if (!meta.exists()) meta.createNewFile();
    if (meta.isFile()) {
        Writer writer = null;
        try {
          writer = new BufferedWriter(new FileWriter(meta));
          //store next index
          writer.write(Integer.toString(i));
        } catch (IOException ioe) {
          logger.debug(ioe);
          logger.warn("Writing meta file failed at '"+meta.getAbsolutePath()+"'");
        }
        finally {
          try {
            if (writer != null) writer.close();
          } catch (IOException ioe) {
            throw new IOException("Flushing meta file failed at '"
              +meta.getAbsolutePath()+"'", ioe);
          }
        }
      } else {
        logger.warn("Cannot write meta file at '"+meta.getAbsolutePath()+
          "', it exists but is not a file.");
      }
  }
  
  private File getNewFile(String dirName, String prefix) throws IOException {
    return getNewFile(dirName, prefix, false);
  }

  @SuppressWarnings("unchecked")
  private File getNewFile(String dirName, final String prefix, 
    boolean createDir) throws IOException {
    String path = realmPath + File.separator + dirName + File.separator;
    synchronized (getLock(path)) {
      //before attempting reading the index file, we try to re-use the same id
      if (dirName.equals("records") && !prefix.equals("local")) {
        String fileName = path + File.separator + prefix;
        File file = new File(fileName);
        if (!file.exists()) return file;
      }
      File meta = new File(path + META_FILE);
      int index = readIndex(meta);
      //in case we get 0 we could:
      // a) be dealing with new empty realm
      // b) realm wihtout a meta-file, in which case start from the last file
      if (index == 0) {
        File dir = new File(path);
        if (!dir.exists()) {
          //containing dir does not exist
          if (createDir) {
            if (!dir.mkdir()) throw new IOException("Cannot create dir at '"
              +dir.getAbsolutePath()+"'");
          } else {
            throw new IOException("Dir '"+dir.getAbsolutePath()+"' does not exist");
          }
        }
        File[] files = dir.listFiles(new FilenameFilter() {
          @Override
          public boolean accept(File dir, String name) {
            return name.startsWith(prefix);
          }
        });
        Arrays.sort(files, new NaturalOrderComparator());
        if (files.length > 0) {
          String name = files[files.length-1].getName();
          if (name.startsWith(prefix)
            && name.substring(prefix.length()).matches("-([0-9])+")) {
            index = Integer.parseInt(name.substring(prefix.length()+1));
          }          
        }
      }
      String name = prefix + "-" + index;
      File candid = new File(path + name);
      while (candid.exists()) {
        index++;
        name = prefix + "-" + index;
        candid = new File(path + name);
      }
      candid.createNewFile();
      storeIndex(meta, index+1);
      return candid;
    }
  }

  @Override
  public Layer writeOverrideRecord(Layer override) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    String baseId = override.getWorldId();
    if (baseId == null || baseId.equals("")) baseId = "local";
    FileWriter fw = null;
    File out = null;
    try {
      out = getNewFile("records", escapeId(baseId));
    } catch (IOException ioe) {
      throw new InternalIOException("Cannot create file for record at '"+realmPath+"'", ioe);
    }
    try {
      fw = new FileWriter(out);
      Document envelope = XmlUtils.newDoc("record");
      setRecordType(envelope.getDocumentElement(), rdef);
      override.setId(unescapeId(out.getName()));
      override.setRealmName(rdef.getRealmName());
      override.toDOM(envelope.getDocumentElement(), rdef.getRealmType(), null);
      XmlUtils.serialize(envelope, fw);
    } catch (IOException ioe) {
      throw new InternalIOException("Cannot write record file at '"
        +out.getAbsolutePath()+"'", ioe);
    } catch (TransformerException te) {
      throw new InternalIOException("Record content serialization failed at '"
        +out.getAbsolutePath()+"'", te);
    } finally {
      try {
        if (fw != null) fw.close();
      } catch (IOException io) {
        throw new InternalIOException("Flushing file failed at '"
          +out.getAbsolutePath()+"'", io);
      }
    }
    return override;
  }

  @Override
  public void deleteOverrideRecord(String id) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File rec = new File(realmPath + File.separator + "records" +
      File.separator + escapeId(id));
    if (!rec.exists())
      throw new NotFoundException("No file for record '"
        + rec.getAbsolutePath() + "'");
    if (!rec.delete())
      throw new InternalIOException("Deleting record file failed at '"
        +rec.getAbsolutePath()+"'");
  }

  @Override
  public Pair<List<Layer>,Date> readWorldRecordsForParent(String prId) throws InternalIOException {
    throw new UnsupportedOperationException("World disk caching is not supported, enable in memory caching");
  }

  @Override
  public Layer readWorldRecord(String id) throws InternalIOException {
    throw new UnsupportedOperationException("World disk caching is not supported, enable in memory caching");
  }

  @Override
  public List<Layer> writeWorldRecordsForParent(List<Layer> doc, String prId)
    throws InternalIOException {
    throw new UnsupportedOperationException("World disk caching is not supported, enable in memory caching");
  }

  @Override
  public void deleteWorldRecordsByParentId(final String prId) throws InternalIOException {
    throw new UnsupportedOperationException("World disk caching is not supported, enable in memory caching");
  }

  @Override
  public ParentRealm writeParentRealm(String id, ParentRealm pr)
    throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File recFile = new File(realmPath + File.separator + "parents" +
      File.separator + escapeId(id));
    if (!recFile.exists())
      throw new NotFoundException("Parent realm record not found at '"
        +recFile.getAbsolutePath()+"'");
    FileWriter fw = null;
    try {
      pr.setId(id);
      Document doc = pr.toDOM(null).getOwnerDocument();
      fw = new FileWriter(recFile);
      XmlUtils.serialize(doc, fw);
    } catch (IOException io) {
      throw new InternalIOException("Error while writing parent record at '"
        +recFile.getAbsolutePath()+"'", io);
    } catch (TransformerException te) {
      throw new InternalIOException("Error while serializing parent record at '"
        +recFile.getAbsolutePath()+"'", te);
    } finally {
      try {
        if (fw != null) fw.close();
      } catch (IOException io) {
        throw new InternalIOException("Flushing file failed at '"
          +recFile.getAbsolutePath()+"'", io);
      }
    }
    return pr;
  }
  
  @Override
  public ParentRealm writeParentRealm(ParentRealm pr) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    FileWriter fw = null;
    File recFile = null;
    try {
      recFile = getNewFile("parents", "P");
    } catch (IOException ioe) {
      throw new InternalIOException("Error creating file for parent record at '"
        +realmPath+"'",ioe);
    }
    try {
      pr.setId(unescapeId(recFile.getName()));
      Document doc = pr.toDOM(null).getOwnerDocument();
      fw = new FileWriter(recFile);
      XmlUtils.serialize(doc, fw);
    } catch (IOException io) {
      throw new InternalIOException("Error while writing parent record at '"
        +recFile.getAbsolutePath()+"'", io);
    } catch (TransformerException te) {
      throw new InternalIOException("Error while serializing parent record at '"
        +recFile.getAbsolutePath()+"'", te);
    } finally {
      try {
        if (fw != null) fw.close();
      } catch (IOException io) {
        throw new InternalIOException("Flushing file failed at '"
          +recFile.getAbsolutePath()+"'", io);
      }
    }
    return pr;
  }

  @Override
  public ParentRealm readParentRealm(String prId) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File recFile = new File(realmPath + File.separator + "parents"
      + File.separator + escapeId(prId));
    if (!recFile.exists())
      throw new NotFoundException("Parent realm record does not exist at '"+recFile.getAbsolutePath()+"'");
    try {
      Document doc = XmlUtils.parse(recFile);
      ParentRealm pr = ParentRealm.fromStoredDOM(doc, prId);
      return pr;
    } catch (IOException ioe) {
      throw new InternalIOException("Cannot read parent realm record at '"
        +recFile.getAbsolutePath()+"'", ioe);
    } catch (SAXException se) {
      throw new InternalIOException("Cannot parse parent realm record at '"
        +recFile.getAbsolutePath()+"'", se);
    } catch (MalformedInputException mie) {
      throw new InternalIOException(mie);
    }
  }

  @Override
  public void deleteParentRealm(String prId) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File rec = new File(realmPath + File.separator + "parents" +
      File.separator + escapeId(prId));
    if (!rec.exists())
      throw new NotFoundException("No file for parent realm record '"
        + rec.getAbsolutePath() + "'");
    if (!rec.delete())
      throw new InternalIOException("Deleting parent realm record failed at '"+rec.getAbsolutePath()+"'");
  }

  @Override
  public List<ParentRealm> readParentRealms() throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File recsDir = new File(realmPath + File.separator + "parents");
    List<ParentRealm> parents = new ArrayList<ParentRealm>();
    if (recsDir.isDirectory()) {
      for (File ff : recsDir.listFiles()) {
        if (ff.isFile() && !ff.getName().startsWith(".")) {
          Document doc;
          ParentRealm pr;
          try {
            doc = XmlUtils.parse(ff);
            pr = ParentRealm.fromStoredDOM(doc, unescapeId(ff.getName()));
          } catch (SAXException ex) {
            throw new InternalIOException("Cannot parse record contents at '"
              + ff.getAbsolutePath() + "'", ex);
          } catch (IOException ex) {
            throw new InternalIOException("Cannot read record file at '"
              + ff.getAbsolutePath() + "'", ex);
          } catch (MalformedInputException mie) {
            throw new InternalIOException(mie);
          }
          parents.add(pr);
        }
      }
    } else {
      throw new NotFoundException("Missing 'parents' directory at '" + recsDir.
        getAbsolutePath());
    }
    return parents;
  }
  
  public List<RealmReaderWriter> listRWs(boolean open) throws InternalIOException {
    File baseDir = new File(basePath);
    if (!baseDir.exists() || !baseDir.isDirectory()) {
      throw new InternalIOException("Base storage location does not exists or not a dir at '"+baseDir.getAbsolutePath()+"'");
    }
    File[] dirs = baseDir.listFiles();
    List<RealmReaderWriter> rws = new ArrayList<RealmReaderWriter>(dirs.length);
    for (File realmDir : dirs) {
      if (!realmDir.isDirectory() || realmDir.getName().startsWith("."))
        continue;
      RealmReaderWriter rw = new RealmFileRW(realmDir.getName(), basePath);
      if (open)
        rw.open();
      rws.add(rw);
    }
    return rws;
  }

  @Override
  public List<RealmDefinition> listRealms() throws InternalIOException {
    File baseDir = new File(basePath);
    if (!baseDir.exists() || !baseDir.isDirectory()) {
      throw new InternalIOException("Base storage location does not exists or not a dir at '"+baseDir.getAbsolutePath()+"'");
    }
    File[] dirs = baseDir.listFiles();
    List<RealmDefinition> rdefs = new ArrayList<RealmDefinition>(dirs.length);
    for (File realmDir : dirs) {
      if (realmDir.isDirectory() && !realmDir.getName().startsWith(".")) {
        File rdF = new File(realmDir.getAbsolutePath() 
          + File.separator + "Realm-Definition");
        if (!rdF.isFile()) {
          logger.warn("Possible realm '"+realmDir.getName()+"' is missing Realm-Definition file and will be ignored");
          continue;
        }
        try {
          String realmName = realmDir.getName();
          Document rdD = XmlUtils.parse(rdF);
          RealmDefinition def = RealmDefinition.fromDOM(rdD, realmName);
          rdefs.add(def);
        } catch (SAXException ex) {
          throw new InternalIOException("Cannot parse Realm-Definition for realm "
            + realmDir.getName() + "'", ex);
        } catch (IOException ex) {
          throw new InternalIOException("Cannot read Realm-Definition for realm "
            + realmDir.getName() + "'", ex);
        } catch (MalformedInputException mie) {
          throw new InternalIOException(mie);
        }
      }
    }
    return rdefs;
  }

  private void setRecordType(Element record, RealmDefinition rdef) {
    record.setAttribute("type", rdef.getRealmType());
  }

  @Override
  public boolean isCachedAsParent() throws InternalIOException {
    throw new UnsupportedOperationException("Wolrd disk caching is not supported.");
  }

  @Override
  public Layer updateCachedAsOriginal(Layer original) throws InternalIOException {
    throw new UnsupportedOperationException("World disk caching is not supported.");
  }

  @Override
  public void deleteCachedAsOriginal(String id) throws InternalIOException {
    throw new UnsupportedOperationException("World disk caching is not supported.");
  }
  
  //master records
  
  @Override
  public List<Layer> readMasterRecords()
    throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File recsDir = new File(realmPath + File.separator + "master");
    if (recsDir.isDirectory()) {
      File[] files = recsDir.listFiles();
      List<Layer> recList = new ArrayList<Layer>(files.length);
      for (File ff : files) {
        if (ff.isFile() && !ff.getName().startsWith(".")) {
          Layer layer;
          try {
            Document doc = XmlUtils.parse(ff);
            setRecordType(doc.getDocumentElement(), rdef);
            //enforce ids/type/layerType in case data comes from migration
            //or is not fully valid
            Element master = getFirstElemWithAttr(doc.getDocumentElement(), 
              "layer", "name", "master-override");
            if (master == null)
              throw new MalformedInputException("Record '"
                +unescapeId(ff.getName())+"'is missing 'override' layer at '"
                +ff.getAbsolutePath()+"'");
            layer = Layer.fromStoredDOM(master);
            layer.setId(unescapeId(ff.getName()));
            layer.setRealmName(rdef.getRealmName());
          } catch (SAXException ex) {
            throw new InternalIOException("Cannot parse record contens at '"
              +ff.getAbsolutePath() +"'", ex);
          } catch (IOException ex) {
            throw new InternalIOException("Cannot read record from file at '"
              +ff.getAbsolutePath()+"'", ex);
          } catch (MalformedInputException mie) {
            throw new InternalIOException(mie);
          }
          recList.add(layer);
        }
      }
      return recList;
    } else {
      throw new NotFoundException("Missing master records directory at '" + recsDir.
        getAbsolutePath());
    }
  }

  @Override
  public Layer readMasterRecord(String id) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File recFile = new File(realmPath + File.separator + "master"
      + File.separator + escapeId(id));
    if (!recFile.isFile()) {
      throw new NotFoundException("Cannot find file for record '" + id + "' at '"
        + recFile.getAbsolutePath() +"'");
    }
    try {
      Document doc = XmlUtils.parse(recFile);
      setRecordType(doc.getDocumentElement(), rdef);
      Element master = getFirstElemWithAttr(doc.getDocumentElement(), 
        "layer", "name", "master-override");
      if (master == null) {
        throw new MalformedInputException("Record '"
          +id+"' is missing a 'master-override' layer at '"
          +recFile.getAbsolutePath()+"'");
      }
      Layer layer = Layer.fromStoredDOM(master);
      layer.setId(id);
      layer.setRealmName(rdef.getRealmName());
      return layer;
    } catch (SAXException ex) {
      throw new InternalIOException("Cannot parse record contens at '"
        + recFile.getAbsolutePath() + "'", ex);
    } catch (IOException ex) {
      throw new InternalIOException("Cannot read record from file at '"
        + recFile.getAbsolutePath() + "'", ex);
     } catch (MalformedInputException mie) {
      throw new InternalIOException(mie);
    }
  }

  @Override
  public Layer writeMasterRecord(String id, Layer update) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File out = new File(realmPath + File.separator + "master" +
      File.separator + escapeId(id));
    if (!out.exists())
      throw new NotFoundException("No file for record '"+out.getAbsolutePath()+"'");
    FileWriter fw = null;
    try {
      fw = new FileWriter(out);
      //we will retain the id's so the on-disk records are not confusing
      update.setId(id);
      update.setRealmName(rdef.getRealmName());
      Document envelope = XmlUtils.newDoc("record");
      setRecordType(envelope.getDocumentElement(), rdef);
      update.toDOM(envelope.getDocumentElement(), rdef.getRealmType(), null);
      XmlUtils.serialize(envelope, fw);
    } catch (IOException ioe) {
      throw new InternalIOException("Cannot write record file at '"
        + out.getAbsolutePath() +"'", ioe);
    } catch (TransformerException te) {
      throw new InternalIOException("Record contents serialization failed at '"
        + out.getAbsolutePath() +"'", te);
    } finally {
      try {
        if (fw != null) fw.close();
      } catch (IOException io) {
        throw new InternalIOException("Flushing file failed at '"
          +out.getAbsolutePath()+"'", io);
      }
    }
    return update;
  }

  @Override
  public Layer writeMasterRecord(Layer masterLayer) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    String baseId ="master";
    FileWriter fw = null;
    File out = null;
    try {
      out = getNewFile("master", escapeId(baseId), true);
    } catch (IOException ioe) {
      throw new InternalIOException("Cannot create file for record at '"
        +realmPath+"'/master/", ioe);
    }
    try {
      fw = new FileWriter(out);
      Document envelope = XmlUtils.newDoc("record");
      setRecordType(envelope.getDocumentElement(), rdef);
      masterLayer.setId(unescapeId(out.getName()));
      masterLayer.setRealmName(rdef.getRealmName());
      masterLayer.toDOM(envelope.getDocumentElement(), rdef.getRealmType(), null);
      XmlUtils.serialize(envelope, fw);
    } catch (IOException ioe) {
      throw new InternalIOException("Cannot write record file at '"
        +out.getAbsolutePath()+"'", ioe);
    } catch (TransformerException te) {
      throw new InternalIOException("Record content serialization failed at '"
        +out.getAbsolutePath()+"'", te);
    } finally {
      try {
        if (fw != null) fw.close();
      } catch (IOException io) {
        throw new InternalIOException("Flushing file failed at '"
          +out.getAbsolutePath()+"'", io);
      }
    }
    return masterLayer;
  }

  @Override
  public void deleteMasterRecord(String id) throws InternalIOException {
    if (rdef == null)
      throw new IllegalStateException("Realm R/W is closed or does not allow this operation at '"
        +basePath+"'");
    File rec = new File(realmPath + File.separator + "master" +
      File.separator + escapeId(id));
    if (!rec.exists())
      throw new NotFoundException("No file for record '"
        + rec.getAbsolutePath() + "'");
    if (!rec.delete())
      throw new InternalIOException("Deleting record file failed at '"
        +rec.getAbsolutePath()+"'");
  }

}
